


import json
from typing import Any, Dict, List, Optional
from requests import Response
from tabulate import tabulate

from .utils import api_get,CliResponse
from .auth import require_auth


def display_attacks_results(data: List[Dict[str, Any]]) -> None:
    display_data: List[Dict[str, Any]] = []
    for d in data:
        row = {k: v for k, v in d.items() if k not in ["stacktrace", "submitted_at_unix", "run_at_unix"]}
        display_data.append(row)
    print(tabulate(display_data, headers="keys"))

@require_auth
def get_attacks(access_token: str, json_format: bool = False, attack_id: Optional[str] = None) -> CliResponse:
    url = f"https://api.sandbox.mindgard.ai/api/v1/results/{attack_id}" if attack_id else "https://api.sandbox.mindgard.ai/api/v1/users/experiments"
    res = api_get(url, access_token)
    data: List[Dict[str, Any]] = res.json() if isinstance(res.json(), list) else [res.json()] # TODO - res.json has different strucutre in sinlge vs multi
    # TODO: URGENT: single resource shoudl return id at the top level

    for item in data:
        if "id" in item:
            url_id = item["id"] 
        else:
            url_id = item["meta"]["id"]
        item["url"] = f"https://sandbox.mindgard.ai/r/attack/{url_id}"

    if json_format:
        print(json.dumps(data))
    else:
        print(json.dumps(data, indent=2)) if attack_id else display_attacks_results(data)
    return CliResponse(0)




import json
import sys
import time
from typing import Any, Dict, List, Optional

import requests
from tabulate import tabulate

from .constants import VERSION

from .utils import api_get, api_post, CliResponse

from .auth import require_auth


# TODO: tidy this
def display_test_results(data: List[Dict[str, Any]]) -> None: # TODO: consider color-coded output for risks
    display_data: List[Dict[str, Any]] = []
    for d in data:
        row = {k: v for k, v in d.items() if k not in ["attacks", "updatedAt", "displayName", "source", "isCompleted", "hasFinished"]}
        row["attack_id"] = d["attacks"][0]["id"]
        row["attack_name"] = d["attacks"][0]["attack"]
        row["state"] = d["attacks"][0]["state_message"]
        row["runtime"] = d["attacks"][0]["runtime"] or time.time() - d["attacks"][0]["run_at_unix"]
        row["attack_risk"] = d["attacks"][0]["risk"] if d["attacks"][0]["state_message"] == "Completed" else "TBD"

        display_data.append(row)
        for attack in d["attacks"][1:]:
            display_data.append({
                "attack_id": attack["id"],
                "attack_name": attack["attack"],
                "state": attack["state_message"],
                "runtime": attack["runtime"] or time.time() - attack["run_at_unix"],
                "attack_risk": attack["risk"] if attack["state_message"] == "Completed" else "TBD",
            })
    print(tabulate(display_data, headers="keys"))


def api_get_tests(access_token: str, test_id: Optional[str] = None) -> List[Dict[str, Any]]:
    url = f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}" if test_id else "https://api.sandbox.mindgard.ai/api/v1/assessments?ungrouped=true"
    res = requests.get(url, headers={
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard/{VERSION}"
    })
    res.raise_for_status()
    data: List[Dict[str, Any]] = res.json() if isinstance(res.json(), list) else [res.json()]

    for item in data:
        test_id = item["id"]
        item["url"] = f"https://sandbox.mindgard.ai/r/tests/{test_id}"

    return data

@require_auth
def get_tests(access_token: str, json_format: bool = False, test_id: Optional[str] = None) -> CliResponse:
    data = api_get_tests(access_token, test_id)
    if json_format:
        print(json.dumps(data))
    else:
        display_test_results(data)
    return CliResponse(0)


@require_auth
def run_test(access_token: str, target_name: str, json_format: bool = False, risk_threshold: int = 80) -> CliResponse:
    if not json_format:
        print("Initiating testing...")
    res = api_post("https://api.sandbox.mindgard.ai/api/v1/assessments", access_token, {"mindgardModelName": target_name})
    if json_format:
        print(json.dumps(res.json())) # TODO: include url
        return CliResponse(0)
    else:
        print("Test initiated. Waiting for results...")
        test_id = res.json()["id"]
        completed = False

        test_res = api_get(f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}", access_token)  
        completed = test_res.json()["hasFinished"]
        display_test_results([test_res.json()])

        num_lines = len(test_res.json()["attacks"]) + 2

        while not completed:
            time.sleep(1)

            sys.stdout.write(f"\033[{num_lines}A") 
            sys.stdout.write("\033[J")

            test_res = api_get(f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}", access_token)  
            completed = test_res.json()["hasFinished"]
            display_test_results([test_res.json()])

        risk_score = test_res.json()["risk"]
        if risk_score > risk_threshold:
            print(f"Test completed - risk score {risk_score} above threshold of {risk_threshold}")
            return CliResponse(1)
        else:
            print(f"Test completed - risk score {risk_score} under threshold of {risk_threshold}")
            return CliResponse(0)


from typing import List, Dict, Any
import requests
from .constants import VERSION, API_BASE

class ApiService():
    def get_tests(self, access_token: str) ->  List[Dict[str, Any]]:
        url = f"{API_BASE}/assessments?ungrouped=true"

        res = requests.get(url, headers={
            "Authorization": f"Bearer {access_token}",
            "User-Agent": f"mindgard/{VERSION}"
        })
        
        res.raise_for_status()
        
        data: List[Dict[str, Any]] = res.json()

        # augment the output with the shortened url
        for test in data:
            test_id = test["id"]
            test["url"] = f"https://sandbox.mindgard.ai/r/test/{test_id}"
            for attack in test["attacks"]:
                attack_id = attack["id"]
                attack["url"] = f"https://sandbox.mindgard.ai/r/attack/{attack_id}"

        return data

    def get_test(self, access_token: str, test_id:str) -> Dict[str, Any]:
        url = f"{API_BASE}/assessments/{test_id}"

        res = requests.get(url, headers={
            "Authorization": f"Bearer {access_token}",
            "User-Agent": f"mindgard/{VERSION}"
        })
        
        res.raise_for_status()
        
        data: Dict[str, Any] = res.json()

        data["url"] = f"https://sandbox.mindgard.ai/r/test/{test_id}"
        for attack in data["attacks"]:
            attack_id = attack["id"]
            attack["url"] = f"https://sandbox.mindgard.ai/r/attack/{attack_id}"

        return data

    def submit_test(self, access_token: str, target_name:str) -> Dict[str, Any]:
        url = f"{API_BASE}/assessments"
        post_body = {"mindgardModelName": target_name}
        res = requests.post(url, headers={
            "Authorization": f"Bearer {access_token}",
            "User-Agent": f"mindgard/{VERSION}"
        }, json=post_body)
        res.raise_for_status()
        data: Dict[str, Any] = res.json()
        return data





import argparse
import sys
from typing import List

from .api_service import ApiService
from .list_tests_command import ListTestsCommand
from .run_test_command import RunTestCommand

from .attacks import get_attacks

from .auth import login
from .constants import VERSION
from .tests import get_tests, run_test
from .utils import is_version_outdated, print_to_stderr


def parse_args(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Securing AIs', prog='mindgard', usage='%(prog)s [command] [options]', epilog='Enjoy the program! :)', add_help=True)
    parser.add_argument('--version', action='version', version=f"%(prog)s {VERSION}", help='Show the current version number')

    subparsers = parser.add_subparsers(dest='command', title='commands', description='Use these commands to interact with the Mindgard API')
    subparsers.add_parser('login', help='Login to the Mindgard platform')

    # TODO: think about more streamlined command for running a test
    tests_parser = subparsers.add_parser('tests', help='See the tests you\'ve run.') # TODO: better help text
    tests_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False) # TODO: don't allow this if run comes after
    tests_parser.add_argument('--id', type=str, help='Get the details of a specific test.', required=False)
    
    tests_subparsers = tests_parser.add_subparsers(dest='test_commands', title='test_commands', description='Perform actions against tests')
    tests_run_parser = tests_subparsers.add_parser('run', help='Run a test.')
    
    # TODO: links to view results in the UI for images etc
    tests_run_parser.add_argument('--name', type=str, help='The attack to tests.', required=True, choices=['cfp_faces', 'mistral'])
    tests_run_parser.add_argument('--json', action="store_true", help='Initiate test and return id that can be used to check status.', required=False)
    tests_run_parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False, default=80)

    # TODO: better error message if someone provides an id that is for the wrong resource eg attacks or tests
    attack_parser = subparsers.add_parser('attacks', help='See the attacks you\'ve run.') # TODO: alias single version of plural nouns
    attack_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False)
    attack_parser.add_argument('--id', type=str, help='Get the details of a specific attack.', required=False)

    # from here is new command structure which we're incrementally adding
    test_parser = subparsers.add_parser('test', help='Test a model')
    # since this feels nonsensical, here's a link: https://docs.python.org/3/library/argparse.html#nargs
    test_parser.add_argument('target', nargs='?', type=str)
    test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    test_parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False, default=80)

    list_parser = subparsers.add_parser('list', help='List items')
    list_subparsers = list_parser.add_subparsers(dest='list_command')
    list_test_parser = list_subparsers.add_parser('tests', help='List tests')
    list_test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    list_test_parser.add_argument('--id', type=str, help='Get the details of a specific test.', required=False)

    return parser.parse_args(args)
    


def main() -> None:
    args = parse_args(sys.argv[1:])

    if not (sys.version_info.major == 3 and sys.version_info.minor >= 8):
        print_to_stderr("Python 3.8 or later is required to run the Mindgard CLI.")
        sys.exit(1)

    if new_version := is_version_outdated():
        print_to_stderr(f"New version available: {new_version}. Run 'pip install mindgard --upgrade' to upgrade. Older versions of the CLI may not be actively maintained.")
    
    if args.command == 'login':
        login()
    elif args.command == 'list' and args.list_command == 'tests':
        api_service = ApiService()
        cmd = ListTestsCommand(api_service)
        res = cmd.run(json_format=bool(args.json), test_id=args.id)
        exit(res.code())
    elif args.command == 'test':
        api_service = ApiService()
        cmd = RunTestCommand(api_service)
        res = cmd.run(model_name=args.target, json_format=bool(args.json), risk_threshold=int(args.risk_threshold))
        exit(res.code())
    elif args.command == 'tests':
        if args.test_commands == "run":
            res = run_test(target_name=args.name, json_format=bool(args.json), risk_threshold=int(args.risk_threshold))
            exit(res.code())
        else:
            res = get_tests(json_format=bool(args.json), test_id=args.id)
            exit(res.code())
    elif args.command == 'attacks':
        res = get_attacks(json_format=args.json, attack_id=args.id)
        exit(res.code())
    else:
        print_to_stderr('Hey give us a command. Use list or auth.') # TODO update


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        print_to_stderr(e)
        exit(2)